import matplotlib.pyplot as plt
import numpy as np

from simprotein import run_MC

# functions for genetic algorithm

# randomly create an initial population of size n_population of chains of length chain_length
def create_initpopulation(chain_length, n_population):
    return (np.random.randint(0, high=2, size=(n_population, chain_length)) * 2 - 1)

# compute fitness (either squared error or relative entropy, compared to given reference distribution)
def fitness(types, ref_ree_dist, ref_bin_edges, tag, folder='./', use_srel=False):

    # simulation parameters
    temp = 1.0
    rc = 2.5

    # run simulation
    _, ree = run_MC(5000, temp, rc, types, save_ree_freq=10, save_fluc_freq=-1)

    # compute end-to-end distance distribution
    hist, bin_edges = np.histogram(ree, bins=ref_bin_edges, density=True)
    bin_centers = (bin_edges[:-1]+bin_edges[1:])/2.

    # compute fitness
    if use_srel:
        srel = np.sum(ref_ree_dist * np.log((ref_ree_dist + 1e-3) / (hist + 1e-3))) # add some epsilon to each bin so that ratio is defined
        fitness_val = 1./srel
    else:
        err = hist-ref_ree_dist
        sq_err = np.sum(err * err)
        fitness_val = 1. / sq_err

    # save distribution
    fname = folder + '/opt_reedist_' + tag + '.txt'
    np.savetxt(fname, np.hstack((bin_centers.reshape(-1, 1),
                                 hist.reshape(-1, 1))))
    return fitness_val

# select parents
def select(population, fitnesses):
    
    # get only top 20% of parents
    fitness_threshold = np.percentile(fitnesses, 80)

    # select, weighted by fitness, 2 parents
    p_select = np.copy(fitnesses)
    p_select[fitnesses < fitness_threshold] = 0
    p_select = p_select / np.sum(p_select)
    n_population = np.shape(population)[0]
    winners = np.random.choice(np.arange(n_population), size=2, p=p_select)

    return population[winners[0], :], population[winners[1], :]

# create two children from two parents
def crossover(parent1, parent2):

    child1 = np.copy(parent1)
    child2 = np.copy(parent2)
    chain_length = len(parent1)

    # randomly select crossover point
    cross_point = np.random.randint(0, high=chain_length)

    # create children
    child1[cross_point:] = parent2[cross_point:]
    child2[cross_point:] = parent1[cross_point:]

    return child1, child2

# mutate a sequence with exponentially decaying rate
def mutation(individual, rate, gen=-1):
    
    # rate decays as generation number increases
    if gen >= 0:
        rate = rate * np.exp(-gen / 100.)

    # mutate by flipping a bead
    for i in range(len(individual)):
        if np.random.random() < rate:
            if individual[i] == -1:
                individual[i] = 1
            else:
                individual[i] = -1

    return individual

# get a reference distribution from an existing end-to-end trajectory (generated by simprotein.py)
def _get_ref_dist(fname_ref_dist=None, bin_edges=None):
    if fname_ref_dist == None:
        fname = 'ree_seq1111111111_T1.0.txt'
    else:
        fname = fname_ref_dist
    if bin_edges is None:
        bin_edges = np.arange(0, 10.5, 0.5)
    ree = np.loadtxt(fname)
    hist, bin_edges = np.histogram(ree, bins=bin_edges, density=True)
    return hist, bin_edges

def run_GA(n_generations, init_pop_size, chain_length, folder_name='.', fname_ref_dist=None):

    mutation_rate = 0.1

    # generate initial population
    population = create_initpopulation(chain_length, init_pop_size)

    ref_ree_dist, ref_bin_edges = _get_ref_dist(fname_ref_dist)    

    # compute fitnesses
    fitnesses = np.array([fitness(individual, ref_ree_dist, ref_bin_edges, str(i), folder=folder_name, use_srel=True) for i, individual in enumerate(population)])

    n_pop = np.shape(population)[0]

    for i_gen in range(n_generations):
        # selection
        parent1, parent2 = select(population, fitnesses)
        
        # crossover
        child1, child2 = crossover(parent1, parent2)
        
        # mutation
        child1 = mutation(child1, mutation_rate, gen=i_gen)
        child2 = mutation(child2, mutation_rate, gen=i_gen)
        
        # add to population
        population = np.append(population, 
                               [child1, child2],
                               axis=0)

        # compute fitness
        fitnesses = np.append(fitnesses,
                              fitness(child1, ref_ree_dist, ref_bin_edges, str(n_pop), folder=folder_name, use_srel=True))
        n_pop += 1
        fitnesses = np.append(fitnesses,
                              fitness(child2, ref_ree_dist, ref_bin_edges, str(n_pop), folder=folder_name, use_srel=True))
        n_pop += 1
        
    fittest_tag = np.argmax(fitnesses)
    return fittest_tag, population, fitnesses

# test the functions
if __name__ == '__main__':
    population = create_initpopulation(10, 10)
    print('population', population)
    
    ref_ree_dist, ref_bin_edges = _get_ref_dist()
    ref_bin_centers = (ref_bin_edges[:-1]+ref_bin_edges[1:])/2.
    plt.plot(ref_bin_centers, ref_ree_dist)
    
    print('fitness for original types', fitness([1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                                                ref_ree_dist, ref_bin_edges, '1', 
                                                use_srel=True))
    
    print('fitness for opposite type', fitness([-1, -1, -1, -1, -1, -1, -1, -1, -1, -1],
                                               ref_ree_dist, ref_bin_edges, '2',
                                               use_srel=True))
    
    print('fitness for middle type', fitness([1, 1, 1, 1, -1, -1, 1, 1, 1, 1],
                                             ref_ree_dist, ref_bin_edges, '3',
                                             use_srel=True))
    
